<?
   // ----------------------------------------
   // Include Files
   // ----------------------------------------

   require('svg.inc');

   // ----------------------------------------
   // Defines
   // ----------------------------------------

   define("SIZE",                1000);
   define("N_CELLS",               75);
   define("OUTPUT_FILE", "seven.html");

   // ----------------------------------------
   // getEmpty
   // ----------------------------------------

   function getEmpty()
   {
      return(-1);
   }

   // ----------------------------------------
   // RC
   // ----------------------------------------

   class RC
   {
      //function __construct($row, $column)
      function RC($row, $column)
      {
         $this->mRow    = $row;
         $this->mColumn = $column;
      }

      var $mRow;
      var $mColumn;
   }

   // ----------------------------------------
   // Board
   // ----------------------------------------

   class Board
   {
      var $mGrid; 

      function __construct($size)
      {
         $this->mGrid = array();

         for($i = 0; $i < $size; $i ++)
         {
            $row = array();

            for($j = 0; $j < $size; $j ++)
               $row[] = getEmpty();

            $this->mGrid[] = $row;
         }
      }

      function getSize()
      {
         return(count($this->mGrid));
      }

      function isFull()
      {
         foreach($this->mGrid as $row)
            foreach($row as $entry)
               if($entry == getEmpty())
                  return(false);

         return(true);
      }

      function getValue($row, $column)
      {
         return($this->mGrid[$row][$column]);
      }

      function dump()
      {
         foreach($this->mGrid as $row)
         {
            foreach($row as $entry)
            {
               $s = "  ";

               if($entry >= 0)
                  $s = sprintf("%2d", $entry);

               print("[" . $s . "]");
            }

            print("\n");
         }
      }

      function getLocations($size)
      {
         $locations = array();
         $row       = 0;

         // Find the starting column number of each place on our board
         // where 'square' could be placed

         for($n = count($this->mGrid), $column = 0; $column < $n; $column ++)
         {
            // If this column's full or if there's not enough space
            // above us to place the square

            if(($row = $this->getRowIndex($column)) == getEmpty() || $row + 1 < $size)
               continue;

            // How many columns after this one also yield that same value for '$row'?

            $count = 1;

            for($i = $column + 1; $i < $n; $i ++)
            {
               if($this->getRowIndex($i) == $row)
                  $count++;

               else
                  break;
            }

            if($count >= $size)
               $locations[] = new RC($row, $column);
         }

         return($locations);
      }

      function place($column, $size, $squareIndex)
      {
         $row = $this->getRowIndex($column);

         for($i = 0; $i < $size; $i ++)
            for($j = 0; $j < $size; $j ++)
               $this->mGrid[$row - $i][$column + $j] = $squareIndex;
      }

      function getRowIndex($column)
      {
         for($i = count($this->mGrid) - 1; $i >= 0; $i --)
            if($this->mGrid[$i][$column] == getEmpty())
               return($i);

         return(getEmpty());
      }

      function render($svg, $squareSizes, $boardSize)
      {
         $xOffset = ($svg->mWidth  - $boardSize) / 2.0;
         $yOffset = ($svg->mHeight - $boardSize) / 2.0;

         $cellSize = (double)$boardSize / $this->getSize();

         $copy = array();

         foreach($this->mGrid as $row)
         {
            $copyRow = array();

            foreach($row as $entry)
               $copyRow[] = $entry;

            $copy[] = $copyRow;
         }

         $cs = count($copy);

         for(;;)
         {
            $found = false;

            for($v = 0, $row = $cs - 1; !$found && $row >= 0; $row --, $v++)
            {
               for($column = 0; !$found && $column < $cs; $column ++)
               {
                  if($copy[$row][$column] >= 0)
                  {
                     $n = $squareSizes[$copy[$row][$column]];
                     $found = true;

                     $x0 = (int)($xOffset + $column * $cellSize + 0.5);
                     $y0 = (int)($yOffset + $v      * $cellSize + 0.5);

                     $x1 = (int)($xOffset + ($column + $n) * $cellSize + 0.5);
                     $y1 = (int)($yOffset + ($v      + $n) * $cellSize + 0.5);

                     $xd = ($x1 - $x0);
                     $yd = ($y1 - $y0);

                     while($x0 + $xd >= $svg->mWidth)
                        $xd --;

                     while($y0 + $yd >= $svg->mHeight)
                        $yd --;

                     $r = rand(170, 255);
                     $g = rand(170, 255);
                     $b = rand(170, 255);

                     $theColor  = sprintf("rgb(%d, %d, %d)", $r, $g, $b);
                     $darkColor = sprintf("rgb(%d, %d, %d)", $r - 75, $g - 75, $b - 75);

                     $svg->rect($x0, $y0, $xd, $yd, SVG_NONE,   $theColor);
                     $svg->rect($x0, $y0, $xd, $yd, $darkColor, SVG_NONE);

                     for($i = 0; $i < $n; $i ++)
                        for($j = 0; $j < $n; $j ++)
                           $copy[$row - $i][$column + $j] = getEmpty();
                  }
               }
            }

            if(!$found)
               break;
         }
      }
   }

   // ----------------------------------------
   // verticalStackingDetected
   // ----------------------------------------

   function verticalStackingDetected($column, $size, $board, $squareSizes)
   {
      // -------------------------------------------------------
      // 1x1 squares are exempt from our "no stacking" rules
      // -------------------------------------------------------

      if($size < 2)
         return(false);

      $row = $board->getRowIndex($column);

      // -----------------------------------------------
      // If there's not enough room under 'row' for
      // 2 squares of our same size, we're done
      // -----------------------------------------------

      if($row + 2 * $size >= $board->getSize())
         return(false);

      // -----------------------------------------------
      // Examine the two squares below the spot
      // where we're about to place our new square
      // -----------------------------------------------

      $theRow = $row;

      for($i = 0; $i < 2; $i ++)
      {
         $theRow += $size;

         $theValue = $board->getValue($theRow, $column);

         if($theValue == getEmpty()) 
            return(false);

         if($squareSizes[$theValue] != $size)
            return(false);

         for($j = 0; $j < $size; $j ++)
            for($k = 0; $k < $size; $k ++)
               if($board->getValue($theRow - $j, $column + $k) != $theValue)
                  return(false);
      }

      return(true);
   }

   // ----------------------------------------
   // horizontalStackingDetected
   // ----------------------------------------

   function horizontalStackingDetected($column, $size, $board, $squareSizes)
   {
      // -------------------------------------------------------
      // 1x1 squares are exempt from our "no stacking" rules
      // -------------------------------------------------------

      if($size < 2)
         return(false);

      // -----------------------------------------------
      // If there's not enough room to our left
      // for 2 squares of our same size, we're done
      // -----------------------------------------------

      if($column < 2 * $size)
         return(false);

      $row = $board->getRowIndex($column);

      $theColumn = $column;

      for($i = 0; $i < 2; $i ++)
      {
         $theColumn -= $size;

         $theValue = $board->getValue($row, $theColumn);

         if($theValue == getEmpty())
            return(false);

         if($squareSizes[$theValue] != $size)
            return(false);

         for($j = 0; $j < $size; $j ++)
            for($k = 0; $k < $size; $k ++)
               if($board->getValue($row - $j, $theColumn + $k) != $theValue)
                  return(false);
      }

      return(true);
   }

   // ----------------------------------------
   // generate
   // ----------------------------------------

   function generate($svg, $size, $sizePixels)
   {
      $board = new Board($size);

      $stop = max(2, $board->getSize() / 10);
 
      $sizes = array();
   
      for($i = 1; $i <= $stop; $i ++)
         $sizes[] = $i;

      $squareSizes = array();
   
      while(!$board->isFull())
      {
         // Create a new, randomly-sized piece
   
         $n = $sizes[rand(0, count($sizes) - 1)];
   
         $locations = $board->getLocations($n);

         if(count($locations) > 0)
         {
            $index = -1;
   
            for($i = 0, $nn = count($locations); $i < $nn; $i ++)
               if($index == -1 || $locations[$i]->mRow > $locations[$index]->mRow)
                  $index = $i;
   
            $column = $locations[$index]->mColumn;
   
            if(!verticalStackingDetected($column,   $n, $board, $squareSizes) &&
               !horizontalStackingDetected($column, $n, $board, $squareSizes))
            {
               $squareSizes[] = $n;
               $board->place($column, $n, count($squareSizes) - 1);
            }
         }
      }
   
      $board->render($svg, $squareSizes, $sizePixels);
   }

   // ----------------------------------------------------------
   // Script Begins Here
   // ----------------------------------------------------------

   $svg = new SVG(SIZE, SIZE);

   generate($svg, N_CELLS, SIZE);

   if(($handle = @fopen(OUTPUT_FILE, "w")) == false)
      die("Could not open output file: " . OUTPUT_FILE . "\n");

   fprintf($handle,"%s\n", SVG::toHTML($svg));

   fclose($handle);

   printf("Successfully wrote: %s\n", OUTPUT_FILE);
?>
